// ============================================================================
// Project:	Deductive Database
// Filename:	str.h
// Purpose:	The String Datatype
// Last Change:	04.08.2017
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 1998-2017 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


/**
 * \file  str.h
 * \brief Defines type \c str_t and some useful string functions
 */


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	STR_INCLUDED
#define STR_INCLUDED

//=============================================================================
// Exported Types:
//=============================================================================

//-----------------------------------------------------------------------------
// str_t:
//-----------------------------------------------------------------------------

/**
 * Type for pointers to constant strings (compatible with string literals).
 */

typedef	const char *str_t;


//-----------------------------------------------------------------------------
// STR_NULL:
//-----------------------------------------------------------------------------

/**
 * Typed NULL pointer for strings.
 */

#define STR_NULL (static_cast<str_t>(0))


//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// str_len:
//-----------------------------------------------------------------------------

/**
 * Determine string length.
 */

int	str_len(register str_t str);

//-----------------------------------------------------------------------------
// str_cpy:
//-----------------------------------------------------------------------------

/**
 * Safe string copy.
 *
 * This is basically the standard C library \c strcpy function,
 * but it has an additional parameter for the buffer length
 * and ensures that never more characters are written to the buffer.
 */

void	str_cpy(register char* buf, int len, register str_t str);

//-----------------------------------------------------------------------------
// str_cat:
//-----------------------------------------------------------------------------

/**
 * Safe string concatenation.
 *
 * This is basically the standard C library \c strcat function,
 * but it has an additional parameter for the buffer length
 * and ensures that never more characters are written to the buffer.
 */

void	str_cat(register char* buf, int len, register str_t str);

//-----------------------------------------------------------------------------
// str_eq:
//-----------------------------------------------------------------------------

/**
 * Compare two strings for equality.
 *
 * This is often a better alternative to the standard C library \c strcmp
 * function, which returns 0 if the two strings are equal.
 * The value might be interpreted as boolean \c false,
 * which would be wrong.
 * This function returns the expected boolean value:
 * \c true if the strings are equal, and \c false if not.
 */

bool	str_eq(register str_t str1, register str_t str2);

//-----------------------------------------------------------------------------
// str_contains:
//-----------------------------------------------------------------------------

/**
 * Check whether character appears in string.
 */

bool	str_contains(register char c, register str_t str);

//-----------------------------------------------------------------------------
// str_indent:
//-----------------------------------------------------------------------------

/**
 * Returns string of spaces of a given length (at most 75):
 */

str_t str_indent(int n);

//-----------------------------------------------------------------------------
// str_error:
//-----------------------------------------------------------------------------

/**
 * System Error Message for an Error Number.
 */

void	str_error(int err_no, char *buf, int buf_size);

//-----------------------------------------------------------------------------
// STR_ERRMSG_LEN:
//-----------------------------------------------------------------------------

/**
 * Suggested Length of Buffer for System Error Messages.
 */

#define STR_ERRMSG_LEN	80

//-----------------------------------------------------------------------------
// str_uint:
//-----------------------------------------------------------------------------

/**
 * Convert unsigned int to text and returns the number of characters.
 */

int str_uint(char *buf, int buf_len, unsigned int n);

//-----------------------------------------------------------------------------
// str_int:
//-----------------------------------------------------------------------------

/**
 * Convert int to text and returns the number of characters.
 */

int str_int(char *buf, int buf_len, int n);

//-----------------------------------------------------------------------------
// STR_INT_LEN:
//-----------------------------------------------------------------------------

/**
 * Suggested Length of Buffer for int Values.
 */

#define STR_INT_LEN	11

//-----------------------------------------------------------------------------
// str_fill:
//-----------------------------------------------------------------------------

/**
 * Returns string of spaces to make integer a given width.
 *
 * Note that \c field_width can be at most 12.
 */

str_t str_fill(int n, int field_width);

//-----------------------------------------------------------------------------
// str_ptr:
//-----------------------------------------------------------------------------

/**
 * Convert pointer to text (hex number) and returns number of characters.
 */

int str_ptr(char *buf, int buf_len, const void *ptr);

//-----------------------------------------------------------------------------
// STR_PTR_LEN:
//-----------------------------------------------------------------------------

/**
 * Suggested Length of Buffer for pointer Values.
 */

#define STR_PTR_LEN	17

//=============================================================================
// End of File:
//=============================================================================

#endif

