// ============================================================================
// Project:	Deductive Database
// Filename:	bench_tcff_5.cpp
// Purpose:	Bottom-Up Evaluation for tc(X,Y), Ver 5: BAM
// Last Change:	19.09.2017
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2017 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// This Benchmark is optional:
//=============================================================================

#include "../base/ver.h"
#if	VER_BENCH_TCFF_5


//=============================================================================
// Interface:
//=============================================================================

#include "../base/ver.h"
#include "../base/str.h"
#include "../base/err.h"
#include "../base/check.h"
#include "../bds/stack.h"
#include "../rel/list_2.h"
#include "../rel/cur_2.h"
#include "../rel/set_2.h"
#include "../rel/rel_n_n.h"
#include "../rel/cur_n_n.h"
#include "../pred/pred.h"
#include "../facts/load.h"
#include "../emu/bam.h"
#include "out.h"

#include "bench_tcff_5.h"


//=============================================================================
// Constructor, Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor (initializes subobjects):
//-----------------------------------------------------------------------------

bench_tcff_5_c::bench_tcff_5_c() :

		// Superclass:
		bench_c(),

		// Abstract machine:
		bam("BAM"),

		// Relations for EDB body literals:
		par_ff("par_ff"),
		par_fb("par_fb"),

		// Relations for duplicate check (IDB predicates):
		tc_bb("tc_bb (DupCheck)"),

		// Result:
		result_list("answer_ff"),
		result_cursor(&result_list)

{
	// No error occurred so far:
	Error = false;

	// Set Magic number:
	CHECK_CODE(Magic = BENCH_TCFF_5_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

bench_tcff_5_c::~bench_tcff_5_c()
{
	// Check this object:
	CHECK_VALID("bench_tcff_5_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Public Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// load: Load data file with EDB facts.
//-----------------------------------------------------------------------------

bool bench_tcff_5_c::load(str_t filename) {

	// Check parameter:
	CHECK(filename != STR_NULL, "bench_tcff_5_c::load: filename is null");

	// Check object:
	CHECK_VALID("bench_tcff_5_c::load");

	// Loader object (will contain data about predicates):
	load_c loader;

	// Predicate (input to loader):
	pred_c par_p("par", ARGTYPE_INT, ARGTYPE_INT);
	par_p.add_rel(&par_ff);
	par_p.add_rel(&par_fb, BIND_FB);

	// Add predicate definition to loader:
	loader.add(&par_p);

	// Load data file:
	return loader.load(filename);
}

//-----------------------------------------------------------------------------
// eval: Do bottom-up evaluation and read result over cursor interface.
//-----------------------------------------------------------------------------

long bench_tcff_5_c::eval()
{

	// number of answers found:
	long num_results = 0;

	// Open Cursor (this does the real work):
	open();

	// Loop while results are found:
	while(fetch()) {

		// Get results:
		int ans_1 = c1();
		int ans_2 = c2();

		// Check whether results are ok:
		if(ans_1 < 0)
			return -1;
		if(ans_2 < 0)
			return -1;

		// Increment number of answers found:
		num_results++;

		// Print the result if below limit:
		out_c::answer_2(num_results, ans_1, ans_2);

	// End of fetch loop:
	}

	// Check for errors:
	if(error())
		return -1;

	// Close cursor:
	close();

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size(&par_ff);
		out_c::rel_size(&par_fb);
		out_c::rel_size(&tc_bb);
		out_c::rel_size(&result_list);
		out_c::rel_size_done();
	}

	// Return number of result tuples:
	return num_results;
}


//-----------------------------------------------------------------------------
// Open cursor (note: load() must be called before this method).
//-----------------------------------------------------------------------------

void bench_tcff_5_c::open()
{
	// --------------------------------------------------------------------
	// Transformed program:
	//
	// 1: answer(A,B) :- tc(A,B).
	// 2: tc(A,B) :- par(A,B).
	// 3: tc(A,B) :- par(A,C), tc(C,B).
	// --------------------------------------------------------------------


	// Initialize BAM:
	bam_prog();

	// Execute BAM program:
	bam.run(0);

	// Open cursor over result:
	result_cursor.open();
}


//-----------------------------------------------------------------------------
// file: List of data files.
//-----------------------------------------------------------------------------

str_t bench_tcff_5_c::file(int i) const
{
	switch(i) {
		case 0:
			return VER_FILE_TC_TEST;
		case 1:
			return VER_FILE_TC_50K;
		case 2:
			return VER_FILE_TC_50K_CYC;
		case 3:
			return VER_FILE_TC_250K;
		case 4:
			return VER_FILE_TC_250K_CYC;
		case 5:
			return VER_FILE_TC_500K;
		case 6:
			return VER_FILE_TC_500K_CYC;
		case 7:
			return VER_FILE_TC_500K_D2;
		case 8:
			return VER_FILE_TC_500K_D2_CYC;
		case 9:
			return VER_FILE_TC_1M_D2;
		case 10:
			return VER_FILE_TC_1M_D2_CYC;
		default:
			return STR_NULL;
	}
}


//-----------------------------------------------------------------------------
// result: List of correct results of eval method.
//-----------------------------------------------------------------------------

long bench_tcff_5_c::result(int i) const
{
	switch(i) {
		case 0: return 20;
			// tc(1,2). tc(1,3). tc(1,20). tc(1,21). tc(1,22).
			// tc(2,3). tc(2,20). tc(2,21). tc(2,22).
			// tc(4,5). tc(4,6).
			// tc(5,6).
			// tc(10,2). tc(10,3). tc(10,20). tc(10,21). tc(10,22).
			// tc(20,21). tc(20,22).
			// tc(21,22).
			//
			//  1: answer(1, 2)
			//  2: answer(2, 3)
			//  3: answer(1, 3)
			//  4: answer(10, 3)
			//  5: answer(4, 5)
			//  6: answer(5, 6)
			//  7: answer(4, 6)
			//  8: answer(10, 2)
			//  9: answer(2, 20)
			// 10: answer(1, 20)
			// 11: answer(10, 20)
			// 12: answer(20, 21)
			// 13: answer(2, 21)
			// 14: answer(1, 21)
			// 15: answer(10, 21)
			// 16: answer(21, 22)
			// 17: answer(20, 22)
			// 18: answer(2, 22)
			// 19: answer(1, 22)
			// 20: answer(10, 22)

		case 1: return -1;
		case 2: return -1;
		case 3: return -1;
		case 4: return -1;
		case 5: return -1;
		case 6: return -1;
		case 7: return -1;
		case 8: return -1;
		case 9: return -1;
		case 10: return -1;
		default: return -1;
	}
}


//=============================================================================
// Auxiliary Method:
//=============================================================================


//-----------------------------------------------------------------------------
// bam_prog: Initialize abstract machine with the Transitive Closure program.
//-----------------------------------------------------------------------------

bool bench_tcff_5_c::bam_prog()
{
	// Number of variables that are used in this program:
	bam.vars(2);
	int x = 0;
	int y = 1;

	// Create relation names:
	//int par_ff_str =
	bam.string("par_ff");
	//int par_fb_str =
	bam.string("par_fb");
	//int tc_bb_str =
	bam.string("tc_bb");
	//int answer_ff_str =
	bam.string("answer_ff");
	if(bam.error())
		return false;

	// Store relations:
	int par_ff_rel = bam.list_2(&par_ff);
	int par_fb_rel = bam.mmap_1_1(&par_fb);
	int tc_bb_rel = bam.set_2(&tc_bb);
	int answer_ff_rel = bam.list_2(&result_list);
	if(bam.error())
		return false;

	// Create cursor for looping over "par(X,Y)" in rule 1:
	int par_x_y_rule1 = bam.cur_2(par_ff_rel);
	if(bam.error())
		return false;

	// Create cursor for looping over "par(X,Y)" in rule 2:
	int par_x_y_rule2 = bam.cur_1_1(par_fb_rel);
	if(bam.error())
		return false;

	// Procedure to start query evaluation:
	// Rule "tc(X, Y) :- par(X, Y)."
	int loop_1_head = bam.next_code_addr();
	bam.loop_list_2(par_x_y_rule1);
	int loop_1_body = bam.next_code_addr();
	bam.get_list_2_col_1(par_x_y_rule1, x);
	bam.get_list_2_col_2(par_x_y_rule1, y);
	int call_tc = bam.next_code_addr();
	bam.call_forward();
	bam.end_loop_list_2(par_x_y_rule1, loop_1_body);
	int loop_1_done = bam.next_code_addr();
	bam.loop_patch(loop_1_head, loop_1_done);
	bam.halt();

	// Procedure for new tc facts, arguments are in variables (x,y):
	int proc_rule2 = bam.next_code_addr();
	bam.call_patch(call_tc, proc_rule2);
	bam.dupcheck_2(tc_bb_rel, x);
	bam.insert_list_2_v(answer_ff_rel, x);
	// Block for recursive rule "tc(X, Z) :- par(X, Y), *tc(Y, Z)":
	// Values are in (x,y), i.e.: tc(new_x, y) :- par(new_x, x), *tc(x, y)"
	// First save variable and cursor that will be changed
	// (and are still needed when returning from this recursive call):
	bam.save_var(x);
	bam.save_mmap_1_1_cur(par_x_y_rule2);
	// Loop over par(X,Y) / par(new_x, x) with given value for Y / x:
	int loop_2_head = bam.next_code_addr();
	bam.loop_mmap_1_1(par_x_y_rule2, BAM_NULL_ADDR, x);
	int loop_2_body = bam.next_code_addr();
	bam.get_mmap_1_1_out_1(par_x_y_rule2, x);
	bam.call(proc_rule2);
	bam.end_loop_mmap_1_1(par_x_y_rule2, loop_2_body);
	int loop_2_done = bam.next_code_addr();
	bam.loop_patch(loop_2_head, loop_2_done);
	// Restore variable and cursor and return to caller:
	bam.restore_mmap_1_1_cur(par_x_y_rule2);
	bam.restore_var(x);
	bam.ret();

	// Stop BAM:
	bam.halt();

	// Dump data of abstract machine including generated code:
	// bam.dump();

	// Return true if data structures and program were successfully stored:
	return bam.error();
}

//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#if	VER_DEBUG
str_t	bench_tcff_5_c::check() const
{
	// Check magic number:
	if(Magic != BENCH_TCFF_5_MAGIC)
		return "wrong magic number";

	// Check sub-objects (relations for EDB predicates):
	CHECK_SUBOBJ(par_ff, "bench_tcff_5_c::check");
	CHECK_SUBOBJ(par_fb, "bench_tcff_5_c::check");
	CHECK_SUBOBJ(tc_bb, "bench_tcff_5_c::check");
	CHECK_SUBOBJ(result_list, "bench_tcff_5_c::check");
	CHECK_SUBOBJ(result_cursor, "bench_tcff_5_c::check");

	// Check sub-objects (abstract machine):
	CHECK_SUBOBJ(bam, "bench_tcff_5_c::check");

	// Ok:
	return STR_NULL;
}

#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#if	VER_DUMP
void bench_tcff_5_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "tc(X,Y) Benchmark Object, Ver. 2";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(bench_tcff_5_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in all EDB relations:
	err_c::dump_str("Number of rows in EDB relations:  ");
	err_c::dump_nl();
	err_c::dump_str("    par_ff: ");
	err_c::dump_int(par_ff.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    par_fb: ");
	err_c::dump_int(par_fb.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in IDB relations (duplicate check):
	err_c::dump_str("Number of rows in IDB relations:");
	err_c::dump_nl();
	err_c::dump_str("    tc_bb:  ");
	err_c::dump_int(tc_bb.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in the result:
	err_c::dump_str("Number of rows in Result:");
	err_c::dump_nl();
	err_c::dump_str("    result_list: ");
	err_c::dump_int(result_list.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

	// Error status:
	err_c::dump_str("Error checking:");
	err_c::dump_nl();
	err_c::dump_str("    Error: ");
	err_c::dump_bool(Error);
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif


//=============================================================================
// End of Conditional Compilation of this Benchmark (VER_BENCH_TCFF_5):
//=============================================================================

#endif

