// ============================================================================
// Project:	Deductive Database
// Filename:	test_set2.h
// Purpose:	A test method for set implementations with two int arguments.
// Last Change:	11.08.2017
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2016-2017 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


/**
 * \file
 * \brief A test method for set implementations with two integer arguments.
 */


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	TEST_SET2_INCLUDED
#define TEST_SET2_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "../base/ver.h"
#endif

#ifndef STR_INCLUDED
#include "../base/str.h"
#endif

#ifndef CHECK_INCLUDED
#include "../base/check.h"
#endif

#ifndef RELSIZE_INCLUDED
#include "../rel/relsize.h"
#endif

#ifndef GEN_CUR2_INCLUDED
#include "gen_cur2.h"
#endif

#ifndef RTEST_INCLUDED
#include "rtest.h"
#endif

//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// TEST_SET2_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long TEST_SET2_MAGIC = 0x5453320AL;	// 'TS2\n'


//=============================================================================
// Class Template for Testing Sets with two Integer Arguments:
//=============================================================================

template<class T> class test_set2_c : public rtest_c {

	public:

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

	test_set2_c(T set, int test_id) :
		rtest_c(set),
		DataCur(data_rel(test_id))
	{
		// Check parameters:
		CHECK_PAR(set, "test_set2_c::constructor");
		CHECK(test_id >= 0 && test_id <= 4,
			"test_set2_c::constructor: unknown test_id");

		// Initialize attributes:
		Set = set;
		TestID = test_id;

		// Set Magic number:
		CHECK_CODE(Magic = TEST_SET2_MAGIC);
	}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

	~test_set2_c()
	{

		// Check this object:
		CHECK_VALID("test_set2_c::destructor");

		// Make this object invalid:
		CHECK_CODE(Magic = 0);
	}

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	test_set2_c(const test_set2_c& obj);		// Not implemented
	test_set2_c& operator=(const test_set2_c& obj);	// Not implemented

//-----------------------------------------------------------------------------
// insert: Insert tuples into set.
//-----------------------------------------------------------------------------

	int insert()
	{
		// Count number of tuples inserted:
		int tuple_no = 0;

		// Open generator for tuples:
		DataCur.open();

		// Insert tuples:
		while(DataCur.fetch()) {
			tuple_no++;

			int c1 = DataCur.c1();
			int c2 = DataCur.c2();
			bool ok = Set->insert(c1, c2);
			if(!ok) {
				DataCur.close();
				rtest_c::inserted_rows(tuple_no);
				rtest_c::error_row_open();
				rtest_c::error_row_int(c1);
				rtest_c::error_row_int(c2);
				rtest_c::error_row_close();
				return tuple_no;
			}
		}

		// Close cursor:
		DataCur.close();

		// Remember number of inserted rows:
		rtest_c::inserted_rows(tuple_no);

		// Ok:
		return 0;
	}

//-----------------------------------------------------------------------------
// lookup: Lookup all tuples that were inserted into the set.
//-----------------------------------------------------------------------------

	int lookup()
	{
		// Count number of tuples that were found:
		int tuple_no = 0;

		// Open generator for tuples:
		DataCur.open();

		// Lookup tuples:
		while(DataCur.fetch()) {
			tuple_no++;

			int c1 = DataCur.c1();
			int c2 = DataCur.c2();
			bool ok = Set->contains(c1, c2);
			if(!ok) {
				DataCur.close();
				rtest_c::error_row_open();
				rtest_c::error_row_int(c1);
				rtest_c::error_row_int(c2);
				rtest_c::error_row_close();
				return tuple_no;
			}
		}

		// Close cursor:
		DataCur.close();

		// Ok:
		return 0;
	}


//-----------------------------------------------------------------------------
// Values of size measures for the set implementation:
//-----------------------------------------------------------------------------

	int size_value(relsize_t relsize) {
		return Set->size_value(relsize);
	}


//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#if	VER_DEBUG
	public:
	// Integrity check:
	str_t	check() const
	{
		// Check magic number:
		if(Magic != TEST_SET2_MAGIC)
			return "wrong magic number";

		// Check superclass:
		str_t msg = rtest_c::check();
		if(msg)
			return msg;

		// Ok:
		return 0;
	}


	private:
	// Magic number (identifies objects of this class for debugging).
	long	Magic;		// Must be "RTEST_MAGIC".
#endif

#if	VER_DUMP
	public:
	// Display data structure:
	virtual void dump(str_t headline = STR_NULL) const
	{
		// Headline:
		if(headline == STR_NULL)
			headline =
			"Test for sets with one int column (test_set2_c)";

		// Call dump method of superclass:
		rtest_c::dump(headline);
	}
#endif

//=============================================================================
// Private Object Members:
//=============================================================================

	private:

	// Set (Relation) to test:
	T Set;

	// ID of the test:
	int TestID;

	// Generator for tuples to insert:
	gen_cur2_c DataCur;

//=============================================================================
// Auxiliary Functions:
//=============================================================================

	private:

//-----------------------------------------------------------------------------
// data_rel: Select tuple generator for initial insertion depending on test ID.
//-----------------------------------------------------------------------------

	int data_rel(int test_id)
	{
		// Check parameter:
		CHECK(test_id >= 0 && test_id <= 4,
			"test_set2_c::data_rel: unknown test_id");

		switch(test_id) {
			case 0:
				// Small sequential relation:
				return 0;
			case 1:
				// One million tuples, sequential:
				return 1;
			case 2:
				// One million tuples, a bit more "random":
				return 2;
			case 3:
				// 100 million tuples, a bit more "random":
				return 3;
			case 4:
				// One billion tuples, a bit more "random":
				return 4;
			default:
				// Invalid test id:
				CHECK_IMPOSSIBLE(
					"test_set2_c::data_rel: bad id");
				return 1;
		}
	}


//=============================================================================
// End of Class Template:
//=============================================================================

};


//=============================================================================
// End of Include File:
//=============================================================================

#endif

